use super::Keyfork;
use crate::openpgp_card::factory_reset_current_card;
use clap::{Args, Parser, Subcommand};
use std::{collections::HashSet, fs::File, io::IsTerminal, path::PathBuf};

use card_backend_pcsc::PcscBackend;

use keyfork_derive_openpgp::{
    openpgp::{
        self,
        armor::{Kind, Writer},
        packet::{signature::SignatureBuilder, UserID},
        policy::StandardPolicy,
        serialize::Marshal,
        types::{KeyFlags, SignatureType},
        Cert,
    },
    XPrv,
};
use keyfork_derive_path_data::paths;
use keyfork_derive_util::DerivationIndex;
use keyfork_mnemonic::Mnemonic;
use keyfork_prompt::{
    default_handler, prompt_validated_passphrase,
    validators::{SecurePinValidator, Validator},
    Message,
};

use keyfork_shard::{openpgp::OpenPGP, Format};

#[derive(thiserror::Error, Debug)]
#[error("Invalid PIN length: {0}")]
pub struct PinLength(usize);

type Result<T, E = Box<dyn std::error::Error>> = std::result::Result<T, E>;

fn derive_key(seed: [u8; 64], index: u8) -> Result<Cert> {
    let subkeys = vec![
        KeyFlags::empty().set_certification(),
        KeyFlags::empty().set_signing(),
        KeyFlags::empty()
            .set_transport_encryption()
            .set_storage_encryption(),
        KeyFlags::empty().set_authentication(),
    ];

    let subkey = DerivationIndex::new(u32::from(index), true)?;
    let path = paths::OPENPGP_SHARD.clone().chain_push(subkey);
    let xprv = XPrv::new(seed)
        .expect("could not construct master key from seed")
        .derive_path(&path)?;
    let userid = UserID::from(format!("Keyfork Shard {index}"));
    let cert = keyfork_derive_openpgp::derive(xprv, &subkeys, &userid)?;
    Ok(cert)
}

#[derive(Subcommand, Clone, Debug)]
pub enum WizardSubcommands {
    GenerateShardSecret(GenerateShardSecret),
    BottomsUp(BottomsUp),
}

/// Create a 256 bit secret and shard the secret to smart cards.
///
/// Smart cards will need to be plugged in periodically during the wizard, where they will be
/// factory reset and provisioned to `m/pgp'/shrd'/<share index>`. The secret can then be recovered
/// with `keyfork recover shard` or `keyfork recover remote-shard`. The share file will be printed
/// to standard output.
#[derive(Args, Clone, Debug)]
pub struct GenerateShardSecret {
    /// The minimum amount of keys required to decrypt the secret.
    #[arg(long)]
    threshold: u8,

    /// The maximum amount of shards.
    #[arg(long)]
    max: u8,

    /// The amount of smart cards to provision per-shard.
    #[arg(long, default_value = "1")]
    keys_per_shard: u8,

    /// The file to write the generated shard file to.
    #[arg(long)]
    output: Option<PathBuf>,

    /// The file to write generated certificates to.
    #[arg(long)]
    cert_output: Option<PathBuf>,
}

/// Create a 256 bit secret and shard the secret to previously known OpenPGP certificates,
/// deriving the default OpenPGP certificate for the secret.
///
/// This command was purpose-built for DEFCON and is not intended to be used normally, as it
/// implies keys used for sharding have been generated by a custom source.
#[derive(Args, Clone, Debug)]
pub struct BottomsUp {
    /// The location of OpenPGP certificates to use when sharding.
    key_discovery: PathBuf,

    /// The minimum amount of keys required to decrypt the secret.
    #[arg(long)]
    threshold: u8,

    /// The file to write the generated shard file to.
    #[arg(long)]
    output_shardfile: PathBuf,

    /// The file to write the generated OpenPGP certificate to.
    #[arg(long)]
    output_cert: PathBuf,

    /// The User ID for the generated OpenPGP certificate.
    #[arg(long, default_value = "Disaster Recovery")]
    user_id: String,
}

impl WizardSubcommands {
    // dispatch
    fn handle(&self) -> Result<()> {
        match self {
            WizardSubcommands::GenerateShardSecret(gss) => gss.handle(),
            WizardSubcommands::BottomsUp(bu) => bu.handle(),
        }
    }
}

fn cross_sign_certs(certs: &mut [Cert]) -> Result<(), Box<dyn std::error::Error>> {
    let policy = StandardPolicy::new();

    #[allow(clippy::unnecessary_to_owned)]
    for signing_cert in certs.to_vec() {
        let mut certify_key = signing_cert
            .with_policy(&policy, None)?
            .keys()
            .unencrypted_secret()
            .for_certification()
            .next()
            .expect("certify key unusable/not found")
            .key()
            .clone()
            .into_keypair()?;
        for signable_cert in certs.iter_mut() {
            let sb = SignatureBuilder::new(SignatureType::GenericCertification);
            let userid = signable_cert
                .userids()
                .next()
                .expect("a signable user ID is necessary to create web of trust");
            let signature = sb.sign_userid_binding(
                &mut certify_key,
                signable_cert.primary_key().key(),
                &userid,
            )?;
            let changed;
            (*signable_cert, changed) = signable_cert.clone().insert_packets2(signature)?;
            assert!(
                changed,
                "OpenPGP certificate was unchanged after inserting packets"
            );
        }
    }
    Ok(())
}

impl GenerateShardSecret {
    fn handle(&self) -> Result<()> {
        let root_entropy = keyfork_entropy::generate_entropy_of_const_size::<{ 256 / 8 }>()?;
        let mnemonic = Mnemonic::from_array(root_entropy);
        let seed = mnemonic.generate_seed(None);
        let mut pm = default_handler()?;
        let mut certs = vec![];
        let mut seen_cards: HashSet<String> = HashSet::new();
        let stdout = std::io::stdout();
        if self.output.is_none() {
            assert!(
                !stdout.is_terminal(),
                "not printing shard to terminal, redirect output"
            );
        }

        let user_pin_validator = SecurePinValidator {
            min_length: Some(6),
            ..Default::default()
        }
        .to_fn();
        let admin_pin_validator = SecurePinValidator {
            min_length: Some(8),
            ..Default::default()
        }
        .to_fn();

        for index in 0..self.max {
            let cert = derive_key(seed, index)?;
            for i in 0..self.keys_per_shard {
                pm.prompt_message(Message::Text(format!(
                    "Please remove all keys and insert key #{} for user #{}",
                    (i as u16) + 1,
                    (index as u16) + 1,
                )))?;
                let card_backend = loop {
                    if let Some(c) = PcscBackend::cards(None)?.next().transpose()? {
                        break c;
                    }
                    pm.prompt_message(Message::Text(
                        "No smart card was found. Please plug in a smart card and press enter"
                            .to_string(),
                    ))?;
                };
                let user_pin = prompt_validated_passphrase(
                    &mut *pm,
                    "Please enter the new smartcard User PIN: ",
                    3,
                    &user_pin_validator,
                )?;
                let admin_pin = prompt_validated_passphrase(
                    &mut *pm,
                    "Please enter the new smartcard Admin PIN: ",
                    3,
                    &admin_pin_validator,
                )?;
                factory_reset_current_card(
                    &mut |application_identifier| {
                        if seen_cards.contains(&application_identifier) {
                            // we were given the same card, error
                            // we're gonna panic because this is a significant error
                            panic!("Previously used card {application_identifier} was reused");
                        } else {
                            seen_cards.insert(application_identifier);
                            true
                        }
                    },
                    user_pin.trim(),
                    admin_pin.trim(),
                    &cert,
                    &openpgp::policy::NullPolicy::new(),
                    card_backend,
                )?;
            }
            certs.push(cert);
        }

        cross_sign_certs(&mut certs)?;

        let opgp = OpenPGP;

        if let Some(output_file) = self.output.as_ref() {
            let output = File::create(output_file)?;
            opgp.shard_and_encrypt(
                self.threshold,
                certs.len() as u8,
                mnemonic.as_bytes(),
                &certs[..],
                output,
            )?;
        } else {
            opgp.shard_and_encrypt(
                self.threshold,
                certs.len() as u8,
                mnemonic.as_bytes(),
                &certs[..],
                std::io::stdout(),
            )?;
        }

        if let Some(cert_output_file) = self.cert_output.as_ref() {
            let output = File::create(cert_output_file)?;
            let mut writer = Writer::new(output, Kind::PublicKey)?;
            for cert in certs {
                cert.serialize(&mut writer)?;
            }
            writer.finalize()?;
        }

        Ok(())
    }
}

impl BottomsUp {
    fn handle(&self) -> Result<()> {
        let entropy = keyfork_entropy::generate_entropy_of_const_size::<{ 256 / 8 }>()?;
        let mnemonic = Mnemonic::from_array(entropy);
        let seed = mnemonic.generate_seed(None);

        // TODO: should this allow for customizing the account index from 0? Potential for key reuse
        // errors.
        let path = paths::OPENPGP_DISASTER_RECOVERY
            .clone()
            .chain_push(DerivationIndex::new(0, true)?);
        let subkeys = [
            KeyFlags::empty().set_certification(),
            KeyFlags::empty().set_signing(),
            KeyFlags::empty()
                .set_transport_encryption()
                .set_storage_encryption(),
            KeyFlags::empty().set_authentication(),
        ];
        let xprv = XPrv::new(seed)
            .expect("could not construct master key from seed")
            .derive_path(&path)?;
        let userid = UserID::from(self.user_id.as_str());

        let cert = keyfork_derive_openpgp::derive(xprv, &subkeys, &userid)?;
        let certfile = File::create(&self.output_cert)?;
        let mut w = Writer::new(certfile, Kind::PublicKey)?;
        cert.serialize(&mut w)?;
        w.finalize()?;

        let opgp = OpenPGP;
        let certs = OpenPGP::discover_certs(&self.key_discovery)?;

        let shardfile = File::create(&self.output_shardfile)?;
        opgp.shard_and_encrypt(
            self.threshold,
            certs.len() as u8,
            &entropy,
            &certs[..],
            shardfile,
        )?;
        Ok(())
    }
}

#[derive(Parser, Debug, Clone)]
pub struct Wizard {
    #[command(subcommand)]
    command: WizardSubcommands,
}

impl Wizard {
    pub fn handle(&self, _k: &Keyfork) -> Result<()> {
        self.command.handle()?;
        Ok(())
    }
}
